/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.ev3;

import org.jabotics.robot.defs.*;
import lejos.hardware.ev3.LocalEV3;
import lejos.hardware.Power;
import lejos.hardware.port.Port;
import lejos.hardware.port.TachoMotorPort;


/**
 * AMotor is the implementation class for the interface IMotorAdapter on the EV3 robot.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class AMotor implements IMotorAdapter
{
    static private AMotor[] sMotors = null;
    static private volatile boolean sTachoCountsAreFrozen = false;

    // Instanzvariablen:
    protected TachoMotorPort mMotorPort;
    private volatile float mTachoCount;
    private volatile float mTachoCountCorrectionFactor;

    static public synchronized AMotor getMotor(EMotorPort port)
    {
        if (null == sMotors) sMotors = new AMotor[] { new AMotor(EMotorPort.A), new AMotor(EMotorPort.B), new AMotor(EMotorPort.C) };

        return sMotors[port.ordinal()];
    }

    private static synchronized void freezeTachoCounts()
    {
        if (!sTachoCountsAreFrozen)
        {
            sTachoCountsAreFrozen = true;
            int n = 0;
            long start;
            long end = System.nanoTime();
            do
            {
                start = end;
                for (AMotor motor: sMotors) motor.mTachoCount = motor.mTachoCountCorrectionFactor * (float)motor.mMotorPort.getTachoCount();
                end = System.nanoTime();
            }            
            while ( (end-start > 1000000) && (++n < 5) );
        }
    }

    private static synchronized void releaseTachoCounts()
    {
        sTachoCountsAreFrozen = false;
    }

    private AMotor(EMotorPort port)
    {
        mMotorPort = LocalEV3.get().getPort(port.name()).open(TachoMotorPort.class);
        mMotorPort.resetTachoCount();
        mMotorPort.setPWMMode(TachoMotorPort.PWM_FLOAT);
        mTachoCount = mMotorPort.getTachoCount();
        mTachoCountCorrectionFactor = 1f;
    }

    private void updateTachoCount()
    {
        synchronized(AMotor.class)
        {
            if (!sTachoCountsAreFrozen)  mTachoCount = mTachoCountCorrectionFactor * (float)mMotorPort.getTachoCount();
        }
    }

    @Override
    public float getMaxSpeed()
    {
        return LocalEV3.get().getPower().getVoltage() * 100.0f;
    }

    @Override
    public void startControl()
    {
        mMotorPort.setPWMMode(TachoMotorPort.PWM_BRAKE);
    }

    @Override
    public void setToFloatMode()
    {
        mMotorPort.controlMotor(0,TachoMotorPort.FLOAT);
    }

    @Override
    public void setPower(float power)
    {
        int powerInt = Math.round(power);
        mMotorPort.controlMotor(powerInt, (0 == powerInt) ? TachoMotorPort.STOP : TachoMotorPort.FORWARD);
    }

    @Override
    public float getTachoCount()
    {
        updateTachoCount();
        return mTachoCount;
    }

    @Override
    public float[] getPIDParameters()
    {
        return new float[] {4f,0.04f,10f,2f,0.02f,8f}; // P (control), I (control),D  (control), P (halted), I (halted), D (halted)
    }

    @Override
    public void setTachoCountCorrectionFactor(float correctionFactor)
    {
        mTachoCountCorrectionFactor = (correctionFactor < 0.9f) ? 0.9f : ((correctionFactor > 1.1f) ? 1.1f : correctionFactor);
    }

    @Override
    public void freezeAllTachoCounts()
    {
        freezeTachoCounts();
    }

    @Override
    public void releaseAllTachoCounts()
    {
        releaseTachoCounts();
    }

    @Override
    public void commitAllPowerData()
    {
    }
}
