/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 * 
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.pc.util;

import java.nio.ByteBuffer;
import java.util.ArrayList;

/*
 * See general comments in XSharedMemory.java
 * 
 * Usage:
 * 1. Instantiate XSharedData
 * 2. Instantiate data elements (derived from BData like XInt, ...)
 * 3. Invoke create-method on XSharedData instance
 * 4. Invoke read-method (initially and when appropriate) to update in-process representation of the data
 * 5. Invoke write-method after making changes to data elements to update shared memory 
 */
public class XSharedData
{
	private static final int INTEGER_BYTES = Integer.SIZE / 8; // Integer.BYTES defined in 1.8
	private static final int FLOAT_BYTES = Float.SIZE / 8;     // Float.BYTES defined in 1.8

	private final String mName;
    private final boolean mMaster;
    private XSharedMemory mMemory = null;
    private ArrayList<BData> mDataElements = new ArrayList<BData>();
    private ByteBuffer mMappedBuffer = null;
    private ByteBuffer mValueBuffer = null;
    private int mByteSize = 0;
    private byte  mWriteCnt = 0;
    private long mInconsistentReadCnt = 0;

    public XSharedData(String name, boolean is_master)
    {
        mName = name;
        mMaster = is_master;
    }

    synchronized private void add(BData element, int byte_size)
    {
        if (null != mMemory) throw new IllegalStateException("XSharedData.add: Memory already has been instantiated. No data elements can be added.");
        mDataElements.add(element);
        mByteSize += byte_size;
    }

    synchronized public void create()
    {
        if (null != mMemory)  throw new IllegalStateException("XSharedData.create: Memory already has been instantiated.");
        mMemory = new XSharedMemory(mName, mByteSize + 6, mMaster); // 6 bytes for control elements
        mMappedBuffer = mMemory.getBuffer();
        mValueBuffer = ByteBuffer.wrap(new byte[mByteSize]);
    }

    synchronized public void read()
    {
        if (null == mMemory) throw new IllegalStateException("XSharedData.read: Memory has not been instantiated yet.");

        // Repeat procedure as often as necessary to get consistent read
        // (This might be complete overkill - if MappedByteBuffer.load() & force()
        //  really determine the synchronization between different processes.)
        while (true)
        {
            mMemory.read();
            mValueBuffer.rewind();
            mMappedBuffer.rewind();
            mWriteCnt = mMappedBuffer.get();
            mMappedBuffer.get(mValueBuffer.array(), 0, mByteSize);
            mValueBuffer.rewind(); // hash code of buffer depends on buffer position!
            int hashCode = mValueBuffer.hashCode();
            if ((mMappedBuffer.getInt() == hashCode) && (mMappedBuffer.get() == mWriteCnt)) break;
            ++mInconsistentReadCnt;
        }

        mValueBuffer.rewind();
        for (BData element : mDataElements)
        {
            element.get(mValueBuffer);
        }
    }

    synchronized public void write()
    {
        if (null == mMemory) throw new IllegalStateException("XSharedData.write: Memory has not been instantiated yet.");

        mValueBuffer.rewind();
        for (BData element : mDataElements)
        {
            element.put(mValueBuffer);
        }

        mValueBuffer.rewind(); // hash code of buffer depends on buffer position!
        int hashCode = mValueBuffer.hashCode();
        mMappedBuffer.rewind();
        mMappedBuffer.put(++mWriteCnt);
        mMappedBuffer.put(mValueBuffer);
        mMappedBuffer.putInt(hashCode);
        mMappedBuffer.put(mWriteCnt);
        mMemory.write();
    }

    synchronized public long getInconsistentReadCount()
    {
        return mInconsistentReadCnt;
    }



    static public abstract class BData
    {
        BData(XSharedData sd, int byte_size)
        {
            sd.add(this, byte_size);
        }

        protected abstract void get(ByteBuffer buffer);
        protected abstract void put(ByteBuffer buffer);
    }

    static public class XInt extends BData
    {
        int mValue = 0;
        
        public XInt(XSharedData sd)
        {
            super(sd, INTEGER_BYTES);
        }

        public int get()
        {
            return mValue;
        }

        public void set(int value)
        {
            mValue = value;
        }

        @Override
        protected void get(ByteBuffer buffer)
        {
            mValue = buffer.getInt();
        }

        @Override
        protected void put(ByteBuffer buffer)
        {
            buffer.putInt(mValue);
        }
    }

    static public class XFloat extends BData
    {
        float mValue = 0;

        public XFloat(XSharedData sd)
        {
            super(sd, FLOAT_BYTES);
        }

        public float get()
        {
            return mValue;
        }

        public void set(float value)
        {
            mValue = value;
        }

        @Override
        protected void get(ByteBuffer buffer)
        {
            mValue = buffer.getFloat();
        }

        @Override
        protected void put(ByteBuffer buffer)
        {
            buffer.putFloat(mValue);
        }
    }

    static public class XIntArray extends BData
    {
        private final int[] mArray;

        public XIntArray(XSharedData sd, int size)
        {
            super(sd, size * INTEGER_BYTES);
            mArray = new int[size];
        }

        public int size()
        {
            return mArray.length;
        }

        public int get(int index)
        {
            return mArray[index];
        }

        public void get(int[] dst)
        {
            for (int i = 0; i < mArray.length; ++i) dst[i] = mArray[i];
        }

        public void get(int[] dst, int offset, int length)
        {
            for (int i = 0, j = offset; i < mArray.length; ++i) dst[j] = mArray[i];
        }

        public void set(int index, int value)
        {
            mArray[index] = value;
        }

        public void set(int[] values)
        {
            for (int i = 0; i < mArray.length; ++i) mArray[i] = values[i];
        }

        @Override
        protected void get(ByteBuffer buffer)
        {
            for (int i = 0; i < mArray.length; ++i) mArray[i] = buffer.getInt();
        }

        @Override
        protected void put(ByteBuffer buffer)
        {
            for (int i = 0; i < mArray.length; ++i) buffer.putInt(mArray[i]);
        }
    }

    static public class XFloatArray extends BData
    {
        private final float[] mArray;

        private XFloatArray(XSharedData sd, int size)
        {
            super(sd, size * FLOAT_BYTES);
            mArray = new float[size];
        }

        public int size()
        {
            return mArray.length;
        }

        public float get(int index)
        {
            return mArray[index];
        }
        
        public void get(float[] dst)
        {
            for (int i = 0; i < mArray.length; ++i) dst[i] = mArray[i];
        }

        public void get(float[] dst, int offset, int length)
        {
            for (int i = 0, j = offset; i < mArray.length; ++i) dst[j] = mArray[i];
        }

        public void set(int index, float value)
        {
            mArray[index] = value;
        }

        public void set(float[] values)
        {
            for (int i = 0; i < mArray.length; ++i) mArray[i] = values[i];
        }

        @Override
        protected void get(ByteBuffer buffer)
        {
            for (int i = 0; i < mArray.length; ++i) mArray[i] = buffer.getFloat();
        }

        @Override
        protected void put(ByteBuffer buffer)
        {
            for (int i = 0; i < mArray.length; ++i) buffer.putFloat(mArray[i]);
        }
    }
}
