/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.en;


/**
 * <font color="LimeGreen"><b>&bull;</b></font> The interface IMotor provides methods for controlling a motor.
 * The following units apply to parameters and return values of method calls:
 *  rotations angles and positions in ° (degrees),
 *  velocities in °/s,
 *  accelerations in °/s² and
 *  time periods in ms.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */

public interface IMotor
{
    /**
     * <font color="Gold"><b>&bull;</b></font> Sets the correction factor for the motor position to the given value.
     * The accuracy of the motor's position sensor is limited leading to deviations from the desired rotations and
     * motions. In particular this is apparent for straight robot motions which in practice turn out to be slightly
     * curved. A correction factor slightly different from 1 for one of the motors can account for this problem.
     * A correction factor smaller than 1 leads to an increase in observed rotation angles and velocities.
     * A correction factor larger than 1 leads to a reduction in observed rotation angles and velocities.
     * The correction factor is limited to the range from 0.9 to 1.1 which is more than sufficient
     * to correct the mentioned inaccuracies.
     * 
     * @param   correctionFactor   correction factor for the motor position (default value is 1)
     */
    public void setCorrectionFactor(float correctionFactor);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the current angular position of the motor.
     * 
     * @return      the current angular position of the motor (in °)
     */
    public float getPosition();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Resets the angular position of the motor.
     */
    public void resetPosition();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the instantaneous
     * (that is "real") rotational velocity of the motor. In contrast to the target velocity,
     * which is always given as a positive figure, the instantaneous velocity also can be negative.
     * A negative speed corresponds to a backward rotation of the motor.
     * 
     * @return  the instantaneous rotational velocity of the motor (in °/s)
     */
    public float getSpeed();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the currently stored target velocity.
     * The target velocity is always given as a positive figure and does not determine
     * the direction of motion.
     * 
     * @return  currently stored target velocity (in °/s)
     */
    public float getTargetSpeed();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target velocity to the given value.
     * The target velocity is always given as a positive figure and does not determine
     * the direction of motion. If the target velocity is set to 0, the current rotation
     * of the motor will be stopped. Technically, there is a lower limit for the
     * target velocity which is in the range of a few °/s.
     * 
     * @param   speed   the new rotational target velocity (in °/s)
     */
    public void setTargetSpeed(float speed);
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the stored target acceleration.
     * The target acceleration is always given as a positive figure and does not determine
     * the direction of rotation or acceleration.
     * 
     * @return  the target acceleration (in °/s²)
     */
    public float getTargetAcceleration();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target acceleration to the given value.
     * The target acceleration is always given as a positive figure and does not determine
     * the direction of rotation or acceleration. Technically, there is a lower limit for
     * accelerations which is in the range of a few °/s².
     * 
     * @param   acceleration   the new target acceleration (in °/s²)
     */
    public void setTargetAcceleration(float acceleration);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the maximum rotational velocity.
     * The maximum velocity might depend on the battery voltage.
     * 
     * @return  the maximum velocity (in °/s)
     */
    public float getMaxSpeed();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target velocity to its maximum value.
     * The maximum velocity might depend on the battery voltage.
     */
    public void setTargetSpeedToMax();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns the maximum acceleration.
     * 
     * @return  the maximum acceleration (in °/s²)
     */
    public float getMaxAcceleration();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Sets the target acceleration to its maximum value.
     */
    public void setTargetAccelerationToMax();

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts an unlimited rotation of the motor.
     *
     * @param  direction    direction of rotation (see {@link EDirection})
     */
    public void rotate(EDirection direction);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a time limited rotation of the motor.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the rotation or whether it will wait until the rotation has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * motor rotation (see {@link EWaitingMode}).
     * Expectations should not be too high regarding the accuracy of the time measurement.
     * Note that if the motor already is in motion, the minimum rotation time will be the deceleration time
     * which depends on the starting velocity and the target acceleration.
     *
     * @param  period       duration of the motor rotation (in ms)
     * @param  direction    direction of rotation (see {@link EDirection})
     * @param  waitingMode  determines whether program execution will be halted during motor rotation
     */
    public void rotateTime(float period, EDirection direction, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a motor rotation by the given angle.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the rotation or whether it will wait until the rotation has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * motor rotation (see {@link EWaitingMode}).
     * 
     * @param  angle        the rotation angle (in °)
     * @param  waitingMode  determines whether program execution will be halted during motor rotation
     */
    public void rotateByAngle(float angle, EWaitingMode waitingMode);

    /**
     * Replaced by {@link #rotateToPosition(float,EWaitingMode)}.
     */
    @Deprecated
    public void rotateToPostion(float position, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Starts or completes a motor rotation to the given motor position.
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the rotation or whether it will wait until the rotation has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * motor rotation (see {@link EWaitingMode}).
     * @since   1.1.1
     * 
     * @param  position     the target position of the motor (in °)
     * @param  waitingMode  determines whether program execution will be halted during motor rotation
     */
    public void rotateToPosition(float position, EWaitingMode waitingMode);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits for the end of the current motor rotation.
     * Program execution will be continued only after the current motor rotation has been terminated.
     */
    public void waitUntilRotationCompleted();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Stops the current rotation obeying the current target acceleration (here: deceleration).
     * The parameter waitingMode determines whether the program execution will continue immediately after starting
     * the deceleration or whether it will wait until the rotation has been terminated before executing the next
     * program step. In the first case, the following program step(s) will be executed in parallel to the
     * deceleration process (see {@link EWaitingMode}).
     * 
     * @param  waitingMode  determines whether program execution will be halted during motor rotation
     */
    public void stopRotation(EWaitingMode waitingMode);
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Releases the control of the motor.
     * After releasing the control, the motor axis can be freely rotated.
     * For a released motor, monitoring of axis position and velocity will be continued.
     */
    public void releaseControl();
    
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Checks if the motor is currently in motion.
     * 
     * @return  "true" if in motion, "false" if not
     */
    public boolean isMoving();
    
    /**
     * <font color="Gold"><b>&bull;</b></font> Checks if the motor control has been stalled.
     * The controller will automatically release the motor, if the deviation of the motor's position
     * from its target position exceeds a certain limit for a certain period of time.
     * The motor than is put into the stalled state. A stalled motor behaves similar to a released motor.
     * 
     * @return  "true" if motor is stalled, "false" if not
     */
    public boolean isStalled();
}
