/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.compile.*;

import java.awt.Color;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Path2D;

/**
 * Class XRobot implements the physical representation of the robot.
 * Coordinates are given in the usual screen coordinate system (SCS)
 * (x to the right, y downwards, positive angle corresponding to
 * clockwise rotation). This corresponds to a right handed
 * 3D coordinate system with z-direction pointing into the screen.
 * In XPreferences, the SCS will be maintained with the origin
 * in the upper left corner of the scenery.
 * In contrast, the robot API is based on a mathematical
 * coordinate system with its x-direction to the right, y-direction
 * pointing upward, and positive angles corresponding to
 * counter-clockwise rotations ("robot coordinate system" RCS;
 * see core.CPilot, de.IPilot, en.IPilot.). This corresponds
 * to a right handed 3D coordinate system with its z-direction
 * pointing out of the screen.
 * Using the RCS in the physical model of the simulation would easily
 * cause confusion in the 2D graphics implementation. Nevertheless,
 * in the settings panel of the UI the start position of the robot
 * is displayed in the RCS with the origin at the default start position
 * of the scenery.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XRobot extends BModelElement implements IRobotAdapter
{
    // Klassenvariablen:
    static private final float WHEELBASE_DEFAULT = 11.2f;
    static private final float WHEEL_DIAMETER_DEFAULT = 5.6f;
    static private final float EXTENSION_RADIUS = (float)Math.sqrt(13.0 * 13.0 + 4.2 * 4.2);
    static private final float EXTENSION_DIAMETER = 2f * EXTENSION_RADIUS;
    static private final Rectangle2D.Float RECT_BODY = new Rectangle2D.Float(-13f,-4.2f,20f,8.4f);
    static private final Rectangle2D.Float RECT_WHEEL_LEFT = new Rectangle2D.Float(-2.4f,-6.8f,4.8f,2.2f);
    static private final Rectangle2D.Float RECT_WHEEL_RIGHT = new Rectangle2D.Float(-2.4f,4.6f,4.8f,2.2f);
    static private final Rectangle2D.Float RECT_TOUCH_SENSOR_FRONT = new Rectangle2D.Float(6.4f,-4.2f,0.6f,8.4f);
    static private final Rectangle2D.Float RECT_TOUCH_SENSOR_BACK = new Rectangle2D.Float(-13f,-4.2f,0.6f,8.4f);

    static class XDynamicData
    {
        private final DynamicData mRef;
        
        private XDynamicData(DynamicData dd)
        {
            mRef = dd;
        }

        float getRobotX() {return (float)mRef.mRobotX;}
        float getRobotY() {return (float)mRef.mRobotY;}
        float getRobotPhi() {return (float)mRef.mRobotPhi;}
        Path2D.Float getTransformedWallPath() {return mRef.mTransformedWallPath;}
    }

    static private class DynamicData
    {
        static private java.util.concurrent.atomic.AtomicInteger sCount = new java.util.concurrent.atomic.AtomicInteger(0);
        public int mCount = sCount.getAndIncrement();

        public double mRobotX;
        public double mRobotY;
        public double mRobotPhi;
        public XMotor.RotationData mMotorData;
        // performance optimization for light sensor and ultrasonic sensor:
        // memorize transformed wall path, if already calculated
        public volatile Path2D.Float mTransformedWallPath;
        
        public DynamicData()
        {
            if (CS.LOGGING_SIM2)
            {
                if (XLogBuffer.isEnabled())
                {
                    synchronized(XLogBuffer.getMonitor())
                    {
                        XLogBuffer.println("DynamicData(): " + mCount);
                    }
                }
            }
            mRobotX = 600.0;
            mRobotY = 400.0;
            mRobotPhi = 0.0;
            mMotorData = new XMotor.RotationData();
            mTransformedWallPath = null;
        }

        public DynamicData(DynamicData dd)
        {
            if (CS.LOGGING_SIM2)
            {
                if (XLogBuffer.isEnabled())
                {
                    synchronized(XLogBuffer.getMonitor())
                    {
                        if (dd.mCount != mCount - 1)  XLogBuffer.println("DynamicData(dd): " + mCount + "  dd: " + dd.mCount);
                    }
                }
            }
            mRobotX = dd.mRobotX;
            mRobotY = dd.mRobotY;
            mRobotPhi = dd.mRobotPhi;
            mMotorData = dd.mMotorData;
            mTransformedWallPath = dd.mTransformedWallPath;
        }
    }

    static public class TouchSensorValues
    {
        public int mValueFront = 0;
        public int mValueBack = 0;
    }

    // Instanzvariablen
    private volatile boolean mDragging = false;
    private volatile DynamicData mDD;
    private volatile DynamicData mDDGraphics;
    private volatile DynamicData mDDSensors;
    private volatile double mTLastUpdateSensors;
    private volatile double mWheelbase = WHEELBASE_DEFAULT;
    private volatile double mWheelRadius = 0.5 * WHEEL_DIAMETER_DEFAULT;
    private volatile EMotorPort mLeftMotorPort;
    private volatile EMotorPort mRightMotorPort;
    private final Object mDDSensorsMonitor = new Object();
    private final Path2D.Float mWallPath;
    private final XWheel mLeftWheel;
    private final XWheel mRightWheel;
    private final XTouchSensor mTouchSensorFront;
    private final XTouchSensor mTouchSensorBack;
    private final XLightSensor mLightSensorDownward;
    private final XLightSensor mLightSensorForward;
    private final XUltrasonicSensor mUltrasonicSensorForward;
    private final XUltrasonicSensor mUltrasonicSensorBackward;
    private final XUltrasonicSensor mUltrasonicSensorToLeft;
    private final XUltrasonicSensor mUltrasonicSensorToRight;
    private final XButton mEnterButton;
    private final XButton mEscapeButton;
    private final XButton mLeftButton;
    private final XButton mRightButton;
    private final XSoundSensor mSoundSensor;
    private final XDisplay mDisplay;
    private final XSpeaker mSpeaker;

    XRobot(XScene scene)
    {
        super();

        mWallPath = scene.getWallPath();

        // members that will be reinitialized on resetSession will be handled here:
        initVolatileMembers();
        // do not reinitialize the following members on resetSession():
        mWheelbase = WHEELBASE_DEFAULT;
        mWheelRadius = 0.5 * WHEEL_DIAMETER_DEFAULT;
        mLeftMotorPort = EMotorPort.C;
        mRightMotorPort = EMotorPort.B;

        // add all motors for dynamic update (EMotorPort.A first; see XMotor.updateDynamics())
        addChild(XMotor.getMotor(EMotorPort.A));
        addChild(XMotor.getMotor(EMotorPort.B));
        addChild(XMotor.getMotor(EMotorPort.C));

        mLeftWheel = new XWheel(mLeftMotorPort);
        mLeftWheel.setTransform(0f,-5.7f,0f);
        addChild(mLeftWheel);
        mRightWheel = new XWheel(mRightMotorPort);
        mRightWheel.setTransform(0f,5.7f,0f);
        addChild(mRightWheel);
        mTouchSensorFront = XTouchSensor.getTouchSensor(ESensorDirection.FRONT);
        mTouchSensorFront.setTransform(6.7f,0f,0f);
        addChild(mTouchSensorFront);
        mTouchSensorBack = XTouchSensor.getTouchSensor(ESensorDirection.BACK);
        mTouchSensorBack.setTransform(-12.7f,0f,0f);
        addChild(mTouchSensorBack);
        mLightSensorDownward = XLightSensor.getLightSensor(ESensorDirection.DOWNWARD);
        mLightSensorDownward.setTransform(5.4f,0f,0f);
        addChild(mLightSensorDownward);
        mLightSensorForward = XLightSensor.getLightSensor(ESensorDirection.FORWARD);
        mLightSensorForward.setTransform(6f,0f,0f);
        addChild(mLightSensorForward);
        mUltrasonicSensorForward = XUltrasonicSensor.getUltrasonicSensor(ESensorDirection.FORWARD);
        mUltrasonicSensorForward.setTransform(-9.8f,0f,0f);
        addChild(mUltrasonicSensorForward);
        mUltrasonicSensorBackward = XUltrasonicSensor.getUltrasonicSensor(ESensorDirection.BACKWARD);
        mUltrasonicSensorBackward.setTransform(-10.8f,0f,D.PI_F);
        addChild(mUltrasonicSensorBackward);
        mUltrasonicSensorToLeft = XUltrasonicSensor.getUltrasonicSensor(ESensorDirection.TO_LEFT);
        mUltrasonicSensorToLeft.setTransform(-2f,4f,-0.5f*D.PI_F);
        addChild(mUltrasonicSensorToLeft);
        mUltrasonicSensorToRight = XUltrasonicSensor.getUltrasonicSensor(ESensorDirection.TO_RIGHT);
        mUltrasonicSensorToRight.setTransform(-2f,-4f,0.5f*D.PI_F);
        addChild(mUltrasonicSensorToRight);
        // no graphics representation for sound sensor => no transform
        mSoundSensor = XSoundSensor.getSoundSensor(ESensorDirection.NONE);
        addChild(mSoundSensor);

        // button graphics defined in robot coordinates => no transforms
        mEnterButton = XButton.getButton(EButtonType.ENTER);
        addChild(mEnterButton);
        mEscapeButton = XButton.getButton(EButtonType.ESCAPE);
        addChild(mEscapeButton);
        mLeftButton = XButton.getButton(EButtonType.LEFT);
        addChild(mLeftButton);
        mRightButton = XButton.getButton(EButtonType.RIGHT);
        addChild(mRightButton);

        // display graphics defined in robot coordinates => no transform
        mDisplay = XDisplay.getDisplay();
        addChild(mDisplay);

        // no graphics representation for speaker => no transform
        mSpeaker = XSpeaker.getSpeaker();
        addChild(mSpeaker);

        // Build graphics for for this element (without children):
        buildGraphics();
    }

    private void initVolatileMembers()
    {
        mDragging = false;
        XPreferences.Read p = XPreferences.getReadInstance();
        DynamicData dd = new DynamicData();
        dd.mRobotX = p.getRobotStartX();
        dd.mRobotY = p.getRobotStartY();
        dd.mRobotPhi = p.getRobotStartPhi();
        mDD = (hasCollision(dd,null)) ? new DynamicData() : dd;
        setTransform((float)mDD.mRobotX,(float)mDD.mRobotY,(float)mDD.mRobotPhi);
        mDDGraphics = mDD;
        mTLastUpdateSensors = -1.;
        getDDSensors(); // initializes mDDSensor
    }

    private void buildGraphics()
    {
        XGraphicsShape axis = new XGraphicsShape(new Rectangle2D.Float(-0.3f,-5.5f,0.6f,11f));
        axis.setPaint(Color.black);
        axis.setDraw(false);
        axis.setFill(true);
        mGraphicsRep.add(axis);

        XGraphicsShape body = new XGraphicsShape(new Rectangle2D.Float(-13f,-4.2f,20f,8.4f));
        body.setPaint(Color.gray);
        body.setDraw(false);
        body.setFill(true);
        mGraphicsRep.add(body);

        XGraphicsShape brick = new XGraphicsShape(new Rectangle2D.Float(-8f,-3.8f,12f,7.6f));
        brick.setPaint(new Color(244,244,240));
        brick.setDraw(false);
        brick.setFill(true);
        mGraphicsRep.add(brick);
    }

    public void prepareGraphicsUpdate(XDynamicData dd)
    {
        if (CS.LOGGING_SIM2)
        {
            if (XLogBuffer.isEnabled())
            {
                synchronized(XLogBuffer.getMonitor())
                {
                    XLogBuffer.println("prepareGraphicsUpdate:  dd = " + dd.mRef.mCount);
                }
            }
        }
        mDDGraphics = dd.mRef;
        XMotor.prepareGraphicsUpdate(mDDGraphics.mMotorData);
    }

    @Override
    public void updateGraphics()
    {
        DynamicData dd = mDDGraphics;
        if (CS.LOGGING_SIM2)
        {
            if (XLogBuffer.isEnabled())
            {
                synchronized(XLogBuffer.getMonitor())
                {
                    XLogBuffer.println("updateGraphics:         dd = " + dd.mCount);
                }
            }
        }
        setTransform((float)dd.mRobotX,(float)dd.mRobotY,(float)dd.mRobotPhi);
    }

    @Override
    public void setMotors(EMotorPort leftMotorPort, EMotorPort rightMotorPort)
    {
        synchronized(this)
        {
            mLeftMotorPort = leftMotorPort;
            mRightMotorPort = rightMotorPort;
            mLeftWheel.setPort(leftMotorPort);
            mRightWheel.setPort(rightMotorPort);
        }
    }

    @Override
    public float getDefaultWheelDiameter()
    {
        return WHEEL_DIAMETER_DEFAULT;
    }

    @Override
    public float getDefaultWheelbase()
    {
        return WHEELBASE_DEFAULT;
    }

    @Override
    public void setWheelDiameter(float wheel_diameter)
    {
        mWheelRadius = 0.5 * wheel_diameter;
    }

    @Override
    public void setWheelbase(float wheelbase)
    {
        mWheelbase = wheelbase;
    }

    @Override
    public void manageVisitor(BModelVisitor visitor)
    {
        visitor.visitRobot(this);
    }

    public void setPositionAndOrientation(float x, float y, float orientation)
    {
        synchronized(this)
        {
            DynamicData dd = new DynamicData(mDD);
            dd.mRobotX = x;
            dd.mRobotY = y;
            dd.mRobotPhi = restrictedAngle(orientation);
            if (!hasCollision(dd,null))  mDD = dd;
        }
    }

    public XDynamicData getDynamicData()
    {
        return new XDynamicData(mDD);
    }

    public XDynamicData getDDSensors()
    {
        synchronized (mDDSensorsMonitor)
        {
            if ((mDD != mDDSensors) && (XSimulationModel.sTime - mTLastUpdateSensors > 0.005))
            {
                mDDSensors = mDD;
                if (null == mDDSensors.mTransformedWallPath)
                {
                    AffineTransform aft = AffineTransform.getRotateInstance(-mDDSensors.mRobotPhi);
                    aft.translate(-mDDSensors.mRobotX,-mDDSensors.mRobotY);
                    mDDSensors.mTransformedWallPath = new Path2D.Float(mWallPath,aft);
                }
                
                mTLastUpdateSensors = XSimulationModel.sTime;
            }
    
            return new XDynamicData(mDDSensors);
        }
    }

    public void enterDragMode()
    {
        mDragging = true;
    }

    public void leaveDragMode()
    {
        mDragging = false;
    }

    public boolean setDraggedPosition(float x, float y)
    {
        if (!mDragging) return false;

        synchronized(this)
        {
            DynamicData dd = new DynamicData(mDD);
            dd.mRobotX = x;
            dd.mRobotY = y;
            if (hasCollision(dd,null)) return false;
            dd.mMotorData = XMotor.getCurrentData();
            mDD = dd;
            return true;
        }
    }

    public boolean setDraggedOrientation(float orientation)
    {
        if (!mDragging) return false;

        synchronized(this)
        {
            DynamicData dd = new DynamicData(mDD);
            dd.mRobotPhi = restrictedAngle(orientation);
            if (hasCollision(dd,null)) return false;
            dd.mMotorData = XMotor.getCurrentData();
            mDD = dd;
            return true;
        }
    }

    public void updateDynamics()
    {
        if (mDragging) return;

        synchronized(this)
        {
            DynamicData dd = new DynamicData(mDD);
            dd.mMotorData = XMotor.getCurrentData();
            double dsLeft = (dd.mMotorData.getRotAngle(mLeftMotorPort) - mDD.mMotorData.getRotAngle(mLeftMotorPort)) * mWheelRadius;
            double dsRight = (dd.mMotorData.getRotAngle(mRightMotorPort) - mDD.mMotorData.getRotAngle(mRightMotorPort)) * mWheelRadius;
            double ds = 0.5 * (dsLeft + dsRight);
            double dOrientation = (dsLeft - dsRight) / mWheelbase; // orientation angles clockwise
            dd.mRobotPhi = restrictedAngle(mDD.mRobotPhi + dOrientation);
            double meanOrientation = restrictedAngle(mDD.mRobotPhi + 0.5 * dOrientation);
            dd.mRobotX = mDD.mRobotX + ds * Math.cos(meanOrientation);
            dd.mRobotY = mDD.mRobotY + ds * Math.sin(meanOrientation);
            TouchSensorValues sensorValues = new TouchSensorValues();
            if (hasCollision(dd,sensorValues)) // hasCollision will either clear or update dd.mTransformedWallPath
            {
                // Motion of left wheel possible?
                ds = 0.5 * dsLeft;
                dOrientation = dsLeft / mWheelbase; // orientation angles clockwise
                dd.mRobotPhi = restrictedAngle(mDD.mRobotPhi + dOrientation);
                meanOrientation = restrictedAngle(mDD.mRobotPhi + 0.5 * dOrientation);
                dd.mRobotX = mDD.mRobotX + ds * Math.cos(meanOrientation);
                dd.mRobotY = mDD.mRobotY + ds * Math.sin(meanOrientation);
                if (hasCollision(dd,null))
                {
                     // Motion of right wheel possible?
                    ds = 0.5 * dsRight;
                    dOrientation = -dsRight / mWheelbase; // orientation angles clockwise
                    dd.mRobotPhi = restrictedAngle(mDD.mRobotPhi + dOrientation);
                    meanOrientation = restrictedAngle(mDD.mRobotPhi + 0.5 * dOrientation);
                    dd.mRobotX = mDD.mRobotX + ds * Math.cos(meanOrientation);
                    dd.mRobotY = mDD.mRobotY + ds * Math.sin(meanOrientation);
                    if (hasCollision(dd,null))
                    {
                        dd.mRobotPhi = mDD.mRobotPhi;
                        dd.mRobotX = mDD.mRobotX;
                        dd.mRobotY = mDD.mRobotY;
                        dd.mTransformedWallPath = mDD.mTransformedWallPath;
                        mTouchSensorFront.setValue(sensorValues.mValueFront);
                        mTouchSensorBack.setValue(sensorValues.mValueBack);
                    }
                    else
                    {
                        mTouchSensorFront.setValue(0);
                        mTouchSensorBack.setValue(0);
                    }
                }
                else
                {
                    mTouchSensorFront.setValue(0);
                    mTouchSensorBack.setValue(0);
                }
            }
            else
            {
                mTouchSensorFront.setValue(0);
                mTouchSensorBack.setValue(0);
            }
    
            mDD = dd;
        }
    }

    Rectangle2D.Float getEnclosingRect(DynamicData dd)
    {
        return (null == dd) ? new Rectangle2D.Float(-EXTENSION_RADIUS, -EXTENSION_RADIUS, EXTENSION_DIAMETER, EXTENSION_DIAMETER)
            : new Rectangle2D.Float((float)dd.mRobotX - EXTENSION_RADIUS, (float)dd.mRobotY - EXTENSION_RADIUS, EXTENSION_DIAMETER, EXTENSION_DIAMETER);
    }

    private boolean hasCollision(DynamicData dd, TouchSensorValues sensorValues)
    {
        dd.mTransformedWallPath = null;
        // prevent robot from leaving scene area during D&D operations:
        if ( (null == sensorValues) && (!XScene.BOUNDS.contains(dd.mRobotX,dd.mRobotY)) ) return true;

        if (!mWallPath.intersects(getEnclosingRect(dd))) return false;

        AffineTransform aft = AffineTransform.getRotateInstance(-dd.mRobotPhi);
        aft.translate(-dd.mRobotX,-dd.mRobotY);
        dd.mTransformedWallPath = new Path2D.Float(mWallPath,aft);
        if (null == sensorValues)
        {
            return (dd.mTransformedWallPath.intersects(RECT_BODY) || dd.mTransformedWallPath.intersects(RECT_WHEEL_LEFT) || dd.mTransformedWallPath.intersects(RECT_WHEEL_RIGHT));
        }
        else
        {
            if (dd.mTransformedWallPath.intersects(RECT_BODY))
            {
                if (dd.mTransformedWallPath.intersects(RECT_TOUCH_SENSOR_FRONT)) sensorValues.mValueFront = 1;
                if (dd.mTransformedWallPath.intersects(RECT_TOUCH_SENSOR_BACK)) sensorValues.mValueBack = 1;
                return true;
            }

            return (dd.mTransformedWallPath.intersects(RECT_WHEEL_LEFT) || dd.mTransformedWallPath.intersects(RECT_WHEEL_RIGHT));
        }
    }

    private double restrictedAngle(double angle)
    {
        //return ((angle >= D.TWO_PI) || (angle < 0)) ? (angle - D.TWO_PI * Math.floor(angle/D.TWO_PI)) : angle;
        if (angle < 0) return angle + D.TWO_PI;
        if (angle >= D.TWO_PI) return angle - D.TWO_PI;
        return angle;
    }

    @Override
    public ISessionData resetSession()
    {
        initVolatileMembers();
        return this;
    }
}
