/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.EMotorPort;
import org.jabotics.robot.defs.ISessionData;
import org.jabotics.robot.util.XTimedThread;
import java.util.ArrayList;

/**
 * Beschreiben Sie hier die Klasse XSimulationModel.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XSimulationModel implements ISessionData
{
    static class XDynamicsUpdater implements XTimedThread.Runnable
    {
        public interface IHook
        {
            public void postDynamicsUpdate();
        }

        private volatile ArrayList<BModelElement> mElements = new ArrayList<BModelElement>();
        private volatile ArrayList<IHook> mHooks = new ArrayList<IHook>();

        private XDynamicsUpdater()
        {
        }

        @Override
        public void init()
        {
        }

        @Override
        public void runTimed()
        {
            XTimer.notifySimulationIsAlive();

            sControllerHalted = sStopController;
            if (sStopController) return;

            sLastTime = sTime;
            sTime = 1e-9 * (double) (XTimer.nanoTimeRobot());
            sDT = sTime - sLastTime;

            if (sDT > 0) // Simulation might be pausing
            {
                ArrayList<BModelElement> elements = mElements;
                for (BModelElement element : elements)
                {
                    element.updateDynamics();
                }
    
                ArrayList<IHook> hooks = mHooks;
                for (IHook hook : hooks)
                {
                    hook.postDynamicsUpdate();
                }
            }
        }

        @Override
        public void terminate()
        {
            // nothing to do since runTimed() should finish quickly
        }
    
        @Override
        public void shutdown()
        {
            //DEBUG AConsole.close();

            // Fenster schließen?
        }

        public synchronized void add(BModelElement element)
        {
            ArrayList<BModelElement> elements = new ArrayList<BModelElement>(mElements);
            elements.add(element);
            mElements = elements;
        }

        public synchronized void remove(BModelElement element)
        {
            ArrayList<BModelElement> elements = new ArrayList<BModelElement>(mElements);
            elements.remove(element);
            mElements = elements;
        }

        public synchronized void add(IHook hook)
        {
            ArrayList<IHook> hooks = new ArrayList<IHook>(mHooks);
            hooks.add(hook);
            mHooks = hooks;
        }

        public synchronized void remove(IHook hook)
        {
            ArrayList<IHook> hooks = new ArrayList<IHook>(mHooks);
            hooks.remove(hook);
            mHooks = hooks;
        }
    }

    static class XUpdateGraphicsVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.updateGraphics();
        }
    }

    static class XStopSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.stopSession();
        }
    }

    static class XResetSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.resetSession();
        }
    }

    static class XStartSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.startSession();
        }
    }


    // Klassenvariablen
    static private final int DYNAMICS_UPDATE_DELAY = 1;

    // ORDERING IS IMPORTANT!
    // As a side effect, the following line of code will add XTimer to the list of session data in first position:
    static volatile double sTime = 1e-9 * (double) (XTimer.nanoTimeRobot());    // in s
    static volatile double sLastTime = sTime;   // in s
    static volatile double sDT = 0.0;   // in s

    static private final XDynamicsUpdater sDynamicsUpdater = new XDynamicsUpdater();
    static private final XTimedThread sDynamicsUpdateThread = new XTimedThread("Dynamics Update Thread",sDynamicsUpdater);
    static private final XUpdateGraphicsVFC sUpdateGraphicsVFC = new XUpdateGraphicsVFC();
    static private final XScene sScene = new XScene(); // currently no support for switching between different scenes => declare as final
    static private final XRobot sRobot = new XRobot(sScene);
    // for session management:
    static private final XStopSessionVFC sStopSessionVFC = new XStopSessionVFC();
    static private final XResetSessionVFC sResetSessionVFC = new XResetSessionVFC();
    static private final XStartSessionVFC sStartSessionVFC = new XStartSessionVFC();
    static private volatile boolean sStopController = false;
    static private volatile boolean sControllerHalted = false;
    static
    {
        // register objects for dynamics update in right order:
        for (EMotorPort port : EMotorPort.values())
        {
            // Motors have to be registered in this natural order
            // for dynamics update (see implementation of XMotor):
            sDynamicsUpdater.add(XMotor.getMotor(port));
        }
        sDynamicsUpdater.add(sRobot);

        // Vorbereiten und Starten des Steuerungs-Threads:
        sDynamicsUpdateThread.enableTimeDilatation(false);
        sDynamicsUpdateThread.setPriority(Thread.MAX_PRIORITY - 1);
        sDynamicsUpdateThread.setDaemon(true);
        sDynamicsUpdateThread.setTimeouts(DYNAMICS_UPDATE_DELAY,DYNAMICS_UPDATE_DELAY,0);
        sDynamicsUpdateThread.start();

        XSession.getSession().add(new XSimulationModel());
    }

    static public XDynamicsUpdater getDynamicsUpdater()
    {
        return sDynamicsUpdater;
    }

    static void stopDynamicsUpdateThread()
    {
        sDynamicsUpdateThread.terminate();
    }

    static void updateGraphics(XRobot.XDynamicData dd)
    {
        sRobot.prepareGraphicsUpdate(dd);
        sRobot.traverseLeafRoot(sUpdateGraphicsVFC);
    }

    static public XRobot getRobot()
    {
        return sRobot;
    }

    static public XScene getScene()
    {
        return sScene;
    }


    private XSimulationModel()
    {
    }

	// Make sure by special key prefix AAB that session data manager
	// for XSimulationModel is second entry in corresponding map in XSession:
	// Re-initialization needs to happen after that of XTimer and prior to others.
    @Override
    public String getKey()
    {
        return "AAB.XSimulationModel";
    }

    @Override
    public void stopSession()
    {
        try
        {
            sStopController = true;
            while (!sControllerHalted) Thread.sleep(1);
        }
        catch (Exception e) {}
        // traversal root to leaf because sensors rely on robot's dynamic data
        sRobot.traverseRootLeaf(sStopSessionVFC);
        //sScene.traverseRootLeaf(sStopSessionVFC);
    }

    @Override
    public ISessionData resetSession()
    {
        // reset time before resetting the robot!
        // (reset of XTimer ought be done already)
        sTime = 1e-9 * (double) (XTimer.nanoTimeRobot());
        sLastTime = sTime;
        sDT = 0.0;
        // traversal root to leaf because sensors rely on robot's dynamic data
        sRobot.traverseRootLeaf(sResetSessionVFC);
        //sScene.traverseRootLeaf(sResetSessionVFC);
        return this;
    }

    @Override
    public void startSession()
    {
        // traversal root to leaf because sensors rely on robot's dynamic data
        sRobot.traverseRootLeaf(sStartSessionVFC);
        //sScene.traverseRootLeaf(sStartSessionVFC);
        sStopController = false;
        sControllerHalted = false; // do not wait for controller thread to update sControllerHalted
    }
}
