/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.ISessionData;


/**
 * Beschreiben Sie hier die Klasse XTimer.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XTimer
{
    static private class SessionDataManager implements ISessionData
    {
        private boolean mWasPausing = false;

        @Override
        public String getKey()
        {
        	// Make sure by special key prefix AAA that session data manager
        	// for XTimer is first entry in corresponding map in XSession:
        	// Re-initialization of XTimer needs to happen prior to others
        	// because some of them depend on the right robot time. 
            return "AAA.XTimer.SessionDataManager";
        }
    
        @Override
        public void stopSession()
        {
            mWasPausing = mIsPausing;
        }
    
        @Override
        public ISessionData resetSession()
        {
            initVolatileStaticData();
            return this;
        }
    
        @Override
        public void startSession()
        {
            if (!mWasPausing) resumeRobot();

        }
    }

    static private final long MILLION = 1000000;
    /* SUSPENDING FOR ROBOT
     * Idea:
     * Make use of depricated methods Thread.suspend() and Thread.resume() for main robot thread.
     * Implementation of robot controller threads should make sure that no blocking issues occur
     * as long as implementation of robot class does not make unintended system calls.
     * Instead, provided class library should be used (e.g. via ITimerAdapter) and no additional
     * threads should be introduced.
     */
    static private volatile Thread mRobotThread;
    static private volatile boolean mRobotThreadIsSuspended;
    static private volatile boolean mIsPausing;
    static private volatile long mTStart;
    static private volatile long mTStartRobot;
    static private volatile long mTPause;
    static private volatile long mConfirmationPeriod;
    static private volatile long mTNextConfirmation;
    static
    {
        initVolatileStaticData();
        XSession.getSession().add(new SessionDataManager());
    }

    private static void initVolatileStaticData()
    {
        mRobotThread = null;
        mRobotThreadIsSuspended = false;
        mIsPausing = true;
        mTStart = System.nanoTime();
        mTStartRobot = mTStart;
        mTPause = mTStart;
        mConfirmationPeriod = 3000000; // 3ms
        mTNextConfirmation = mTStart + mConfirmationPeriod;
    }


    public static long currentTimeMillis()
    {
        return nanoTime() / MILLION;
    }

    public static long nanoTime()
    {
        return System.nanoTime() - mTStart;
    }

    public static void msDelay(long period)
    {
        nsDelay(period * MILLION);
    }

    public static void nsDelay(long period)
    {
        if (period > 0)
        {
            int nanos = (int)(period % MILLION);
            long millis = (period - nanos) / MILLION;
            try
            {
                Thread.sleep(millis,nanos);
            }
            catch (InterruptedException e)
            {
                Thread.currentThread().interrupt();
            }
        }
    }

    public static long currentTimeMillisRobot()
    {
        return nanoTimeRobot() / MILLION;
    }

    synchronized public static long nanoTimeRobot()
    {
        if (mIsPausing) return getPausedTime();
        long nanoTime = System.nanoTime();
        long exceedingNextConfirmation = nanoTime - mTNextConfirmation;
        if (exceedingNextConfirmation > 0)
        {
            mTNextConfirmation = nanoTime;
            mTStartRobot += exceedingNextConfirmation;
        }
        return nanoTime - mTStartRobot;
    }

    synchronized public static void notifySimulationIsAlive()
    {
        mTNextConfirmation = System.nanoTime() + mConfirmationPeriod;
    }

    public static void msDelayRobot(long period)
    {
        nsDelayRobot(period * MILLION);
    }

    public static void nsDelayRobot(long period)
    {
        long remaining = period;
        long start = nanoTimeRobot();
        while (remaining > 0)
        {
            nsDelay(remaining);
            if (Thread.currentThread().isInterrupted()) return;
            remaining = period - (nanoTimeRobot() - start);
        }
    }

    synchronized private static long getPausedTime()
    {
        return mTPause - mTStartRobot;
    }

    synchronized public static void setRobotThread(Thread thread)
    {
        mRobotThread = thread;
    }

    @SuppressWarnings("deprecation")
    synchronized public static void pauseRobot(boolean suspendThread)
    {
        if (!mIsPausing)
        {
            mTPause = System.nanoTime();
            mIsPausing = true;
    
            /*
             * Use of depricated method Thread.suspend() should work as long as robot program
             * does not make use of unintended system calls and does not introduce additional threads.
             * It should make use of provided methods lika ATimer.msDelay() ...
             * No blocking issues should occur in this case.
             */
            if (suspendThread && (null != mRobotThread))
            {
                mRobotThread.suspend();
                mRobotThreadIsSuspended = true;
            }
        }
    }

    synchronized public static void resumeRobot()
    {
        if (mIsPausing)
        {
            long tCurrent = System.nanoTime();
            mTStartRobot += tCurrent - mTPause;
            mTPause = tCurrent;
            mIsPausing = false;
    
            resumeRobotThread();
        }
    }

    @SuppressWarnings("deprecation")
    synchronized public static void resumeRobotThread()
    {
        if (mRobotThreadIsSuspended)
        {
            mRobotThread.resume();
            mRobotThreadIsSuspended = false;
        }
    }
}