/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import java.awt.Color;
import java.awt.geom.*;

/**
 * Beschreiben Sie hier die Klasse XUltrasonicSensor.
 * 
 * @author  Frank Gollnik
 * @version 1.1.2
 */
public class XUltrasonicSensor extends BSensor
{
    // Konstanten / Klassenvariablen:
    private static final float DETECTION_DISTANCE = 250f;
    private static final float DETECTION_WIDTH = 8f;
    private static final Area DETECTION_AREA = new Area(new Rectangle2D.Float(0f,-0.5f*DETECTION_WIDTH,DETECTION_DISTANCE,DETECTION_WIDTH)); // in sensor's coordinate space
    private static XUltrasonicSensor sUltrasonicSensorForward;
    private static XUltrasonicSensor sUltrasonicSensorBackward;
    private static XUltrasonicSensor sUltrasonicSensorToLeft;
    private static XUltrasonicSensor sUltrasonicSensorToRight;

    static public synchronized XUltrasonicSensor getUltrasonicSensor(ESensorDirection direction)
    {
        if (null == sUltrasonicSensorForward)
        {
            sUltrasonicSensorForward = new XUltrasonicSensor(ESensorDirection.FORWARD);
            sUltrasonicSensorBackward = new XUltrasonicSensor(ESensorDirection.BACKWARD);
            sUltrasonicSensorToLeft = new XUltrasonicSensor(ESensorDirection.TO_LEFT);
            sUltrasonicSensorToRight = new XUltrasonicSensor(ESensorDirection.TO_RIGHT);
        }

        switch (direction)
        {
            case BACKWARD: return sUltrasonicSensorBackward;
            case TO_LEFT:  return sUltrasonicSensorToLeft;
            case TO_RIGHT: return sUltrasonicSensorToRight;
            default:       return sUltrasonicSensorForward;
        }
    }

    // Instanzvariablen:
    private double mLastUpdateTime = -1.;
    private float mValue = 0f;
    private float mCenterX = 0f; // in robot's coordinate space
    private float mCenterY = 0f; // in robot's coordinate space
    private Area mDetectionArea = DETECTION_AREA; // in robot's coordinate space
    private XGraphicsShape mGraphicsTriangle;

    private XUltrasonicSensor(ESensorDirection direction)
    {
        super();
        mDirection = direction;
        buildGraphics();
    }

    private void buildGraphics()
    {
        Path2D.Float path = new Path2D.Float();
        path.moveTo(0f,-2.5f);
        path.lineTo(0f,2.5f);
        path.lineTo(1.0f,0f);
        path.closePath();
        mGraphicsTriangle = new XGraphicsShape(path);
        mGraphicsTriangle.setDraw(false);
        mGraphicsTriangle.setPaint(new Color(51,102,153));
        this.updateGraphics();
        mGraphicsRep.add(mGraphicsTriangle);
    }

    @Override
    public void updateGraphics()
    {
        mGraphicsTriangle.setFill(this.isReferenced());
    }

    @Override
    public ESensorType getType()
    {
        return ESensorType.ULTRASONIC_SENSOR;
    }

    @Override
    public synchronized float getValue(XSensorRef ref)
    {
        if (XSimulationModel.sTime - mLastUpdateTime > 0.006)
        {
            final float distance = getDistance(mDetectionArea, mCenterX, mCenterY, null);
            // no intersection of detection area with walls => distance < 0
            // distance might be slightly larger than DETECTION_DISTANCE due to width of detection area
            mValue = ((distance < 0) || (distance > DETECTION_DISTANCE)) ? DETECTION_DISTANCE : distance;
            mLastUpdateTime = XSimulationModel.sTime;
        }

        return mValue;
    }

    @Override
    public synchronized void setTransform(float translationX, float translationY, float rotation)
    {
        mCenterX = translationX; // in robot's coordinate space
        mCenterY = translationY; // in robot's coordinate space
        AffineTransform atf = AffineTransform.getTranslateInstance(translationX,translationY);
        atf.rotate(rotation);
        mDetectionArea = DETECTION_AREA.createTransformedArea(atf); // in robot's coordinate space

        mGraphicsRep.setTransform(translationX, translationY, rotation);
    }

    @Override
    public int getUpdatePeriodProlongationFactor()
    {
        return 10;
    }

    @Override
    public void manageVisitor(BModelVisitor visitor)
    {
        visitor.visitUltrasonicSensor(this);
    }

    @Override
    public ISessionData resetSession()
    {
        mLastUpdateTime = -1.;
        mValue = 0f;
        return this;
    }
}
